#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "board.h"
#include "random.h"

// X座標
int board_x[BOARD_MAX];

// Y座標
int board_y[BOARD_MAX];

// 4近傍の数
int nb4_empty[PATTERN3_MAX];

// 眼
int eye[PATTERN3_MAX];



// 初期化
struct game *InitializeGame( void )
{
	struct game *game = (struct game *)malloc(sizeof(struct game));

	memset(game, 0, sizeof(struct game));
	ClearGame(game);

	return game;
}


struct game *SetUpGame( void )
{
	struct game *game = (struct game *)malloc(sizeof(struct game));

	memset(game, 0, sizeof(struct game));

	return game;
}



// 初期設定
void ClearGame( struct game *game )
{
	int x, y;
	static struct move_data m = { S_EMPTY, PASS };

	memset(game->prisoner, 0, sizeof(int)*S_MAX);

	game->moves = 1;
	memset(game->record, 0, sizeof(struct move_data)*MAX_RECORDS);

	game->ko = m;
	game->ko_move = -1;

	memset(game->board, 0, sizeof(struct stone_data)*BOARD_MAX);
	for( y=0; y < BOARD_SIZE; y++ ){
		for( x=0; x < OB_SIZE; x++ ){
			game->board[POS(x,y)].color = S_OB;
			game->board[POS(BOARD_SIZE-1-x,y)].color = S_OB;
			game->board[POS(y,x)].color = S_OB;
			game->board[POS(y,BOARD_SIZE-1-x)].color = S_OB;
		}
	}
			
	memset(game->string, 0, sizeof(struct string_data)*MAX_STRINGS);
	game->next_string = 0;
	game->liberty = 0;
	game->liberty_mark = 0;
	game->string_mark = 0;
	memset(game->mark, 0, sizeof(int)*BOARD_MAX);

	ClearPattern3( game->pat3 );

	memset(game->candidates, 0, sizeof(int)*BOARD_MAX);
	for( y=BOARD_START; y<=BOARD_END; y++ ){
		for( x=BOARD_START; x<=BOARD_END; x++ ){
			game->candidates[POS(x,y)] = 1;
		}
		game->sum_candidates_column[y] = PURE_BOARD_SIZE;
	}

	game->sum_candidates = (PURE_BOARD_SIZE * PURE_BOARD_SIZE);
}



// コピー
void CopyGame( struct game *g2, struct game *g1)
{
	 
	memcpy(g2->prisoner, g1->prisoner, sizeof(int)*S_MAX);
	
	g2->moves = g1->moves;
	memcpy(g2->record, g1->record, sizeof(struct move_data)*MAX_RECORDS);

	g2->ko = g1->ko;
	g2->ko_move = g1->ko_move;

	memcpy(g2->board, g1->board, sizeof(struct stone_data)*BOARD_MAX);

	memcpy(g2->string, g1->string, sizeof(struct string_data)*MAX_STRINGS);
	g2->next_string = g1->next_string;
	g2->liberty = g1->liberty;
	g2->liberty_mark = g1->liberty_mark;
	g2->string_mark = g1->string_mark;
	memcpy(g2->mark, g1->mark, sizeof(int)*BOARD_MAX);

	memcpy(g2->pat3, g1->pat3, sizeof(int)*BOARD_MAX);

	memcpy(g2->candidates, g1->candidates, sizeof(int)*BOARD_MAX);
	memcpy(g2->sum_candidates_column, g1->sum_candidates_column, sizeof(int)*BOARD_SIZE);
	g2->sum_candidates = g1->sum_candidates;
}


// メモリ解放
void DoneGame( struct game *game )
{
	if (game) free(game);
}

// 定数の設定
void BoardConst( void )
{
	int x, y;

	memset(board_x, 0, sizeof(board_x));
	memset(board_y, 0, sizeof(board_y));

	for( y=BOARD_START; y<=BOARD_END; y++ ){
		for( x=BOARD_START; x<=BOARD_END; x++ ){
			board_x[POS(x, y)] = x;
			board_y[POS(x, y)] = y;
		}
	}

	InitializeNeighbor();
	InitializeEye();
}
// 近傍の数の初期化
void InitializeNeighbor( void )
{
	int i;
	int empty;

	for (i=0; i<PATTERN3_MAX; i++) {
		empty = 0;

		if (((i>>2) & 0x3) == S_EMPTY) empty++;
		if (((i>>6) & 0x3) == S_EMPTY) empty++;
		if (((i>>8) & 0x3) == S_EMPTY) empty++;
		if (((i>>12) & 0x3) == S_EMPTY) empty++;

		nb4_empty[i] = empty;
	}
}

// 眼の初期化
void InitializeEye( void )
{
	int i, j;
	int transp[8];
	static int eye_pat3[16] = { 
		0x5554, 0x5556, 0x5544, 0x5546, 0x1554,
		0x1556, 0x1544, 0x1546, 0x1564, 0x1146,
		0xFD54, 0xFD55, 0xFD56,
		0xFF74, 0xFF75, 0xFF76
	};

	eye[0x5555] = S_BLACK;
	eye[Pattern3Reverse(0x5555)] = S_WHITE;
	eye[0x1144] = S_BLACK;
	eye[Pattern3Reverse(0x1144)] = S_WHITE;

	for (i=0; i<16; i++) {
		Pattern3Transpose8( eye_pat3[i], transp );
		for (j=0; j<8; j++) {
			eye[transp[j]] = S_BLACK;
			eye[Pattern3Reverse(transp[j])] = S_WHITE;
		}
	}
}

// 碁盤を表示
void DisplayBoard( struct game *game )
{
	static char stone[] = { '+', '@', 'O', '?' };
	int x, y;

	fprintf( stderr, "### Board ### move=%d\n", game->moves );
	fprintf( stderr, "    [");
	for( y=1; y<=PURE_BOARD_SIZE; y++ ){
		fprintf( stderr, "%2d", y);
	}
	fprintf( stderr, "]\n" );
	for( y=BOARD_START; y<=BOARD_END; y++ ){
		fprintf( stderr, "[%2d] ", y-(OB_SIZE-1) );
		for( x=BOARD_START; x<=BOARD_END; x++ ){
			fprintf( stderr, " %c", stone[game->board[POS(x, y)].color] );
		}
		fprintf( stderr, "\n" );
	}
	fprintf( stderr, "\n" );
}

// 合法手判定
int SimpleCheckLegal( struct game *game, int color, int pos )
{
	// 自殺手
	if( nb4_empty[game->pat3[pos]] == 0 && CheckSuicide( game, color, pos) == TRUE ){
		return( FALSE );
	}

	// 劫
	if( game->moves > 1 ){
		if( game->ko.pos == pos && game->ko_move == (game->moves-1) ){
			return( FALSE );
		}
	}

	return( TRUE );
}

int CheckLegal( struct game *game, int color, int pos )
{
	// 空点
	if( game->board[pos].color != S_EMPTY ){
		return( FALSE );
	}

	// 自殺手
	if( nb4_empty[game->pat3[pos]] == 0 && CheckSuicide( game, color, pos) == TRUE ){
		return( FALSE );
	}

	// 劫
	if( game->moves > 1 ){
		if( game->ko.pos == pos && game->ko_move == (game->moves-1) ){
			return( FALSE );
		}
	}

	return( TRUE );
}

// 眼の判定
int CheckEye( struct game *game, int color, int pos )
{
	struct stone_data *board = game->board;
	struct string_data *string = game->string;

	if(eye[game->pat3[pos]] != color
		|| string[board[NORTH(pos)].string_number].liberties == 1	// アタリ
		|| string[board[EAST(pos)].string_number].liberties  == 1
		|| string[board[SOUTH(pos)].string_number].liberties == 1
		|| string[board[WEST(pos)].string_number].liberties  == 1){
			return( FALSE );
	}

	return( TRUE );
}

// 合法手かつ眼でない
int CheckLegalAndEye( struct game *game, int color, int pos )
{
	struct stone_data *board = game->board;
	struct string_data *string = game->string;

	// 眼
	if(eye[game->pat3[pos]] != color
		|| string[board[NORTH(pos)].string_number].liberties == 1	// アタリ
		|| string[board[EAST(pos)].string_number].liberties  == 1
		|| string[board[SOUTH(pos)].string_number].liberties == 1
		|| string[board[WEST(pos)].string_number].liberties  == 1){
			// 自殺手
			if( nb4_empty[game->pat3[pos]] == 0 && CheckSuicide( game, color, pos ) == TRUE ){
				return( FALSE );
			}

			// 劫 
			if( game->moves > 1 ){
				if( game->ko.pos == pos && game->ko_move == (game->moves-1) ){
					return( FALSE );
				}
			}

			return( TRUE );
	}

	return( FALSE );
}

// 自殺手の判定                                                      
int CheckSuicide( struct game *game, int color, int pos )
{
	struct stone_data *board = game->board;
	struct string_data *string = game->string;
	int other = OTHER_COLOR(color);

	// 上
	if( board[NORTH(pos)].color == other ){
		if(string[board[NORTH(pos)].string_number].liberties == 1){
			return( FALSE );
		}
	}else if( board[NORTH(pos)].color == color ){
		if(string[board[NORTH(pos)].string_number].liberties > 1){
			return( FALSE );
		}
	}

	// 右
	if( board[EAST(pos)].color == other){
		if(string[board[EAST(pos)].string_number].liberties == 1){
			return( FALSE );
		}
	}else if( board[EAST(pos)].color == color){
		if(string[board[EAST(pos)].string_number].liberties > 1){
			return( FALSE );
		}
	}

	// 下
	if( board[SOUTH(pos)].color == other ){
		if(string[board[SOUTH(pos)].string_number].liberties == 1){
			return( FALSE );
		}
	}else if( board[SOUTH(pos)].color == color){
		if(string[board[SOUTH(pos)].string_number].liberties > 1){
			return( FALSE );
		}
	}

	// 左
	if( board[WEST(pos)].color == other){
		if(string[board[WEST(pos)].string_number].liberties == 1){
			return( FALSE );
		}
	}else if( board[WEST(pos)].color == color){
		if(string[board[WEST(pos)].string_number].liberties > 1){
			return( FALSE );
		}
	}

	return( TRUE );
}

// 碁盤に石を置く
void SetStone( struct game *game, int color, int pos )
{
	struct stone_data *board = game->board;
	struct string_data *string = game->string;
	int other = OTHER_COLOR(color);
	int prisoner = 0;
	int neighbor_allies = 0;
	int str=0;

	// 棋譜
	game->record[game->moves].color = color;
	game->record[game->moves].pos = pos;

	// パス
	if(pos == PASS){
		// 手数
		game->moves++;
		return;
	}

	// 石を置く
	board[pos].color = color;

	// 候補手
	game->candidates[pos] = 0;
	game->sum_candidates_column[board_y[pos]]--;
	game->sum_candidates--;

	// パターン
	if(color == S_BLACK){
		UpdatePattern3Black( game->pat3, pos );
	}else{
		UpdatePattern3White( game->pat3, pos );
	}

	// マーク更新
	game->string_mark++;

	// 上
	if (board[NORTH(pos)].color == color) {
		neighbor_allies++;
		str = board[NORTH(pos)].string_number;
		string[board[NORTH(pos)].string_number].mark = game->string_mark;
	}else if (board[NORTH(pos)].color == other) {
		if(string[board[NORTH(pos)].string_number].liberties > 1){
			UpdateLiberties( game, board[NORTH(pos)].string_number );
			string[board[NORTH(pos)].string_number].mark = game->string_mark;
		}else{
			prisoner += RemoveStone( game, board[NORTH(pos)].string_number );
		}
	}

	// 右
	if (board[EAST(pos)].color == color && string[board[EAST(pos)].string_number].mark != game->string_mark) {
		neighbor_allies++;
		str = board[EAST(pos)].string_number;
		string[board[EAST(pos)].string_number].mark = game->string_mark;
	}else if (board[EAST(pos)].color == other && string[board[EAST(pos)].string_number].mark != game->string_mark) {
		if(string[board[EAST(pos)].string_number].liberties > 1){
			UpdateLiberties( game, board[EAST(pos)].string_number );
			string[board[EAST(pos)].string_number].mark = game->string_mark;
		}else{
			prisoner += RemoveStone( game, board[EAST(pos)].string_number );
		}
	}

	// 下
	if (board[SOUTH(pos)].color == color && string[board[SOUTH(pos)].string_number].mark != game->string_mark) {
		neighbor_allies++;
		str = board[SOUTH(pos)].string_number;
		string[board[SOUTH(pos)].string_number].mark = game->string_mark;
	}else if (board[SOUTH(pos)].color == other && string[board[SOUTH(pos)].string_number].mark != game->string_mark) {
		if(string[board[SOUTH(pos)].string_number].liberties > 1){
			UpdateLiberties( game, board[SOUTH(pos)].string_number );
			string[board[SOUTH(pos)].string_number].mark = game->string_mark;
		}else{
			prisoner += RemoveStone( game, board[SOUTH(pos)].string_number );
		}
	}

	// 左
	if (board[WEST(pos)].color == color && string[board[WEST(pos)].string_number].mark != game->string_mark) {
		neighbor_allies++;
		str = board[WEST(pos)].string_number;
		string[board[WEST(pos)].string_number].mark = game->string_mark;
	}else if (board[WEST(pos)].color == other && string[board[WEST(pos)].string_number].mark != game->string_mark) {
		if(string[board[WEST(pos)].string_number].liberties > 1){
			UpdateLiberties( game, board[WEST(pos)].string_number );
			string[board[WEST(pos)].string_number].mark = game->string_mark;
		}else{
			prisoner += RemoveStone( game, board[WEST(pos)].string_number );
		}
	}

	// アゲハマ
	game->prisoner[color] += prisoner;

	// 連
	if(neighbor_allies == 0){
		CreateNewString( game, pos );
		// 劫の判定
		str = board[pos].string_number;
		if(prisoner == 1 && string[str].liberties == 1 && string[str].size == 1){
			game->ko_move = game->moves;
			game->ko.color = color;
			game->ko.pos = game->liberty;
		}
	}else if(neighbor_allies == 1){
		ExtendNeighborString( game, str, pos );
	}else{
		AssimilateNeighborString( game, pos );
	}

	// 手数
	game->moves++;
}

// 囲まれた石を取り除く
int RemoveStone( struct game *game, int str )
{
	struct stone_data *board = game->board;
	struct string_data *string = game->string;
	int other = OTHER_COLOR(string[str].color);
	int pos;
	int size = string[str].size;
	int other_str;

	pos = string[str].origin;
	do {
		// パターンの更新
		UpdatePattern3Empty( game->pat3, pos );

		// 初期化
		game->board[pos].color = S_EMPTY;
		game->board[pos].string_number = 0;

		// 候補手
		game->candidates[pos] = 1;
		game->sum_candidates_column[board_y[pos]]++;
		game->sum_candidates++;

		// 周囲のother石のダメ更新
		other_str = board[NORTH(pos)].string_number;
		if(board[NORTH(pos)].color == other
			&& game->mark[other_str] != game->string_mark
			&& other_str != 0){
				UpdateLiberties( game, other_str);
				game->mark[other_str] = game->string_mark;
		}
		other_str = board[EAST(pos)].string_number;
		if(board[EAST(pos)].color == other
			&& game->mark[other_str] != game->string_mark
			&& other_str != 0){
				UpdateLiberties( game, other_str);
				game->mark[other_str] = game->string_mark;
		}
		other_str = board[SOUTH(pos)].string_number;
		if(board[SOUTH(pos)].color == other
			&& game->mark[other_str] != game->string_mark
			&& other_str != 0){
				UpdateLiberties( game, other_str);
				game->mark[other_str] = game->string_mark;
		}
		other_str = board[WEST(pos)].string_number;
		if(board[WEST(pos)].color == other
			&& game->mark[other_str] != game->string_mark
			&& other_str != 0){
				UpdateLiberties( game, other_str);
				game->mark[other_str] = game->string_mark;
		}

		pos = game->board[pos].next_stone;
	} while (pos != string[str].origin);

	// 連を解放する
	string[str].mark = 0;

	// 取った石の数を返す
	return size;
}

// 連の新規作成
void CreateNewString( struct game *game, int pos )
{
	struct stone_data *board = game->board;
	struct string_data *string = game->string;
	int str;

	// 連番号更新
	do{
		game->next_string++;
		if(game->next_string > (MAX_STRINGS-1)) game->next_string = 1;
	}while(string[game->next_string].mark != 0);

	str = game->next_string;
	board[pos].string_number = str;

	board[pos].next_stone = pos;

	// 初期化
	string[str].color = game->board[pos].color;
	string[str].size = 1;
	string[str].origin = pos;
	string[str].liberties = 0;

	// マーク更新
	game->string_mark++;

	// マーキング
	string[str].mark = game->string_mark;

	UpdateLiberties( game, str );
}

// 周囲の連に石を追加
void ExtendNeighborString( struct game *game, int str, int pos )
{
	struct stone_data *board = game->board;
	struct string_data *string = game->string;
	int pos2;

	string[str].size++;

	// 連番号統一
	board[pos].string_number = str;

	// リンク
	pos2 = string[str].origin;
	board[pos].next_stone = board[pos2].next_stone;
	board[pos2].next_stone = pos;

	if (pos < pos2) string[str].origin = pos;

	// ダメの更新
	UpdateLiberties( game, str );
}

// 連の作成 + 同化
void AssimilateNeighborString( struct game *game, int pos )
{
	struct stone_data *board = game->board;
	struct string_data *string = game->string;
	int str;
	int color = board[pos].color;

	// 連番号更新
	do{
		game->next_string++;
		if(game->next_string > (MAX_STRINGS-1)) game->next_string = 1;
	}while(string[game->next_string].mark != 0);
	str = game->next_string;
	board[pos].string_number = str;

	board[pos].next_stone = pos;

	// 初期化
	string[str].color = color;
	string[str].size = 1;
	string[str].origin = pos;
	string[str].liberties = 0;

	// マーク更新
	game->string_mark++;

	// マーキング
	string[str].mark = game->string_mark;

	// 上
	if (board[NORTH(pos)].color == color && string[board[NORTH(pos)].string_number].mark != game->string_mark) {
		AssimilateString( game, str, NORTH(pos) );
	}

	// 右
	if (board[EAST(pos)].color == color && string[board[EAST(pos)].string_number].mark != game->string_mark) {
		AssimilateString( game, str, EAST(pos) );
	}

	// 下
	if (board[SOUTH(pos)].color == color && string[board[SOUTH(pos)].string_number].mark != game->string_mark) {
		AssimilateString( game, str, SOUTH(pos) );
	}

	// 左
	if (board[WEST(pos)].color == color && string[ board[WEST(pos)].string_number].mark != game->string_mark) {
		AssimilateString( game, str, WEST(pos));
	}

	// ダメの更新
	UpdateLiberties( game, str );
}

// 連の同化
void AssimilateString( struct game *game, int str, int pos )
{
	struct stone_data *board = game->board;
	struct string_data *string = game->string;
	int last, pos2;
	int str2 = board[pos].string_number;
	string[str].size += string[str2].size;

	// 連番号統一 
	pos = string[str2].origin;
	do {
		board[pos].string_number = str;
		last = pos;
		pos = board[pos].next_stone;
	} while (pos != string[str2].origin);

	// リンク 
	pos2 = string[str].origin;
	board[last].next_stone = board[pos2].next_stone;
	board[pos2].next_stone = string[str2].origin;

	if (string[str2].origin < pos2){
		string[str].origin = string[str2].origin;
	}

	// 破棄した連番号を解放 
	string[str2].mark = 0;
}

// ダメの更新
void UpdateLiberties( struct game *game, int str )
{
	struct stone_data *board = game->board;
	struct string_data *string = game->string;
	int pos;
	int c;

	// 初期化 
	string[str].liberties = 0;

	// マーク更新 
	game->liberty_mark--;

	pos = string[str].origin;
	do {
		if (board[NORTH(pos)].color == S_EMPTY && game->mark[NORTH(pos)] != game->liberty_mark) {
			game->liberty = NORTH(pos);
			string[str].liberties++;
			if(string[str].liberties >= MAX_LIBERTIES) break;
			game->mark[NORTH(pos)] = game->liberty_mark;
		}

		if (board[EAST(pos)].color == S_EMPTY && game->mark[EAST(pos)] != game->liberty_mark) {
			game->liberty = EAST(pos);
			string[str].liberties++;
			if(string[str].liberties >= MAX_LIBERTIES) break;
			game->mark[EAST(pos)] = game->liberty_mark;
		}

		if (board[SOUTH(pos)].color == S_EMPTY && game->mark[SOUTH(pos)] != game->liberty_mark) {
			game->liberty = SOUTH(pos);
			string[str].liberties++;
			if(string[str].liberties >= MAX_LIBERTIES) break;
			game->mark[SOUTH(pos)] = game->liberty_mark;
		}

		if (board[WEST(pos)].color == S_EMPTY && game->mark[WEST(pos)] != game->liberty_mark) {
			game->liberty = WEST(pos);
			string[str].liberties++;
			if(string[str].liberties >= MAX_LIBERTIES) break;
			game->mark[WEST(pos)] = game->liberty_mark;
		}
		pos = board[pos].next_stone;
	} while (pos != string[str].origin);

	if(string[str].liberties == 1){
		// 候補手 
		c = game->candidates[game->liberty];
		game->candidates[game->liberty] = 1;
		game->sum_candidates_column[board_y[game->liberty]] += (!c);
		game->sum_candidates  += (!c);
	}
}

// 連の表示
void DisplayString( struct game *game )
{
	int x, y, pos;

	printf( "### String ID ###\n" );
	for( y=BOARD_START; y<=BOARD_END; y++ ){
		for( x=BOARD_START; x<=BOARD_END; x++ ){
			pos = POS(x, y);
			if( game->board[pos].string_number == 0 ){
				printf( "   +" );
			}else{
				printf( "%4d", game->board[pos].string_number );
			}
		}
		printf( "\n" );
	}
	printf( "\n" );
}




// 地の計算                                            
int BoardScore( struct game *game )
{
	struct stone_data *board = game->board;
	int x, y;
	int pos;
	int color;
	int scores[S_MAX];

	memset(scores, 0, sizeof(scores));

	for (y=BOARD_START; y<=BOARD_END; y++) {
		for (x=BOARD_START; x<=BOARD_END; x++) {
			pos = POS(x, y);
			color = board[pos].color;
			if (color == S_EMPTY) color = eye[game->pat3[pos]];
			scores[color]++;
		}
	}

	// 黒－白を返す(コミなし) 
	return( scores[S_BLACK] - scores[S_WHITE] );
}