#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <memory.h>
#include <time.h>
#include <float.h>

#include "ucb.h"
#include "board.h"
#include "playout.h"
#include "lower.h"

//UCB値を用いて着手(木構造なし)
int UCBMove( struct game *game, int color )
{
	struct game *po_game = SetUpGame();
	struct pData pd[BOARD_MAX];
	int i, x, y, pos, max_pos, max_count;
	int playout_count = 100000; // やりきれない数として設定
	float ucb_value, max_ucb, win_per;
	clock_t begin_time;
	double finish_time;

	begin_time = clock();
	memset(pd, 0, sizeof( pd ));
	
	for(i = 0; i < playout_count ; ++i){
		max_ucb = 0;
		max_pos = 0;

		for(y = BOARD_START; y <= BOARD_END; ++y){
			for(x = BOARD_START; x <= BOARD_END; ++x){
				pos = POS(x, y);

				if( !game->candidates[pos] || !CheckLegalAndEye( game, color, pos ) ) continue;
				
				// UCB値初期化
				if( !pd[pos].move_count ) ucb_value = FLT_MAX;
				// UCB値計算
				else{
					win_per = (float)pd[pos].win_num/pd[pos].move_count; 
					ucb_value = (float)( win_per + 0.4*sqrt( log( (double) i )/pd[pos].move_count) );
				}
			
				// UCB最大値更新
				if(ucb_value > max_ucb){
					max_ucb = ucb_value;
					max_pos = pos;
				}

			}
		}

		CopyGame(po_game, game);
		SetStone(po_game, color, max_pos);
		pd[max_pos].win_num += 1 - PlayOut( po_game, OTHER_COLOR(color) );
		pd[max_pos].move_count++;

		// 時間制限
		finish_time = GetSpendTime(begin_time);
		if( finish_time >= 0.99 ){ // 念のため0.99秒まで
			playout_count = i;
			break;
		}

	}

	DoneGame(po_game);

	max_count = pd[0].move_count;
	for( i = 1; i < BOARD_MAX; ++i){
		if(pd[i].move_count > max_count){
			max_count = pd[i].move_count;
			max_pos = i;
		}
	}

	// パス
	if( !max_pos ){
		max_pos = PASS;
	}
	// 投了
	else{
		if ( (float)pd[max_pos].win_num/pd[max_pos].move_count <= 0.05 )
			max_pos = RESIGN;
	}

	fprintf(stderr, "%5f sec, (%d games/sec)\n", finish_time, (int)( playout_count/finish_time ));
	fflush(stderr);

	return max_pos;
}